/*
 * Pidgin-libnotify - Provides a libnotify interface for Pidgin
 * Copyright (C) 2005-2007 Duarte Henriques
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "gln_intl.h"

#ifndef PURPLE_PLUGINS
#define PURPLE_PLUGINS
#endif

#include <pidgin.h>
#include <version.h>
#include <debug.h>
#include <util.h>
#include <privacy.h>

/* for pidgin_create_prpl_icon */
#include <gtkutils.h>
#include <gtkblist.h>

#include <libnotify/notify.h>
#include <libindicate/indicator-message.h>
#include <libindicate/server.h>

#include <string.h>

#define PLUGIN_ID "pidgin-libnotify"

#ifdef G_LOG_DOMAIN
#undef G_LOG_DOMAIN
#endif
#define G_LOG_DOMAIN "pidgin-libnotify-plugin"

static GHashTable *buddy_hash;

static gboolean  notify_supports_actions = FALSE;
static gboolean  notify_supports_append = FALSE;
static gboolean  notify_supports_truncation = FALSE;

static gboolean  visibility_managed = FALSE;

static IndicateServer * indicate_server = NULL;

static PurplePluginPrefFrame *
get_plugin_pref_frame (PurplePlugin *plugin)
{
	PurplePluginPrefFrame *frame;
	PurplePluginPref *ppref;

	frame = purple_plugin_pref_frame_new ();

	ppref = purple_plugin_pref_new_with_name_and_label (
                            "/plugins/gtk/libnotify/newmsg",
                            _("New messages"));
	purple_plugin_pref_frame_add (frame, ppref);

	ppref = purple_plugin_pref_new_with_name_and_label (
                            "/plugins/gtk/libnotify/newconvonly",
                            _("Only new conversations"));
	purple_plugin_pref_frame_add (frame, ppref);

	ppref = purple_plugin_pref_new_with_name_and_label (
                            "/plugins/gtk/libnotify/blocked",
                            _("Ignore events from blocked users"));
	purple_plugin_pref_frame_add (frame, ppref);

	ppref = purple_plugin_pref_new_with_name_and_label (
                            "/plugins/gtk/libnotify/signon",
                            _("Buddy signs on"));
	purple_plugin_pref_frame_add (frame, ppref);

	ppref = purple_plugin_pref_new_with_name_and_label (
                            "/plugins/gtk/libnotify/signoff",
                            _("Buddy signs off"));
	purple_plugin_pref_frame_add (frame, ppref);

	ppref = purple_plugin_pref_new_with_name_and_label (
                            "/plugins/gtk/libnotify/only_available",
                            _("Only when available"));
	purple_plugin_pref_frame_add (frame, ppref);

	return frame;
}

/* do NOT g_free() the string returned by this function */
static gchar *
best_name (PurpleBuddy *buddy)
{
	if (buddy->alias) {
		return buddy->alias;
	} else if (buddy->server_alias) {
		return buddy->server_alias;
	} else {
		return buddy->name;
	}
}

static GdkPixbuf *
pixbuf_from_buddy_icon (PurpleBuddyIcon *buddy_icon)
{
	GdkPixbuf *icon;
	const guchar *data;
	size_t len;
	GdkPixbufLoader *loader;

	data = purple_buddy_icon_get_data (buddy_icon, &len);

	loader = gdk_pixbuf_loader_new ();
	gdk_pixbuf_loader_set_size (loader, 48, 48);
	gdk_pixbuf_loader_write (loader, data, len, NULL);
	gdk_pixbuf_loader_close (loader, NULL);

	icon = gdk_pixbuf_loader_get_pixbuf (loader);

	if (icon) {
		g_object_ref (icon);
	}

	g_object_unref (loader);

	return icon;
}

static void
action_cb (NotifyNotification *notification,
		   gchar *action, gpointer user_data)
{
	PurpleBuddy *buddy = NULL;
	PurpleConversation *conv = NULL;

	purple_debug_info (PLUGIN_ID, "action_cb(), "
					"notification: 0x%x, action: '%s'", GPOINTER_TO_UINT(notification), action);

	buddy = (PurpleBuddy *)g_object_get_data (G_OBJECT(notification), "buddy");

	if (!buddy) {
		purple_debug_warning (PLUGIN_ID, "Got no buddy!");
		return;
	}

	conv = purple_find_conversation_with_account (PURPLE_CONV_TYPE_ANY, buddy->name, buddy->account);

	if (!conv) {
		conv = purple_conversation_new (PURPLE_CONV_TYPE_IM,
									  buddy->account,
									  buddy->name);
	}
	conv->ui_ops->present (conv);

	notify_notification_close (notification, NULL);
}

static gboolean
closed_cb (NotifyNotification *notification)
{
	PurpleContact *contact;

	purple_debug_info (PLUGIN_ID, "closed_cb(), notification: 0x%x\n", GPOINTER_TO_UINT(notification));

	contact = (PurpleContact *)g_object_get_data (G_OBJECT(notification), "contact");
	if (contact)
		g_hash_table_remove (buddy_hash, contact);

	g_object_unref (G_OBJECT(notification));

	return FALSE;
}

/* you must g_free the returned string
 * num_chars is utf-8 characters */
static gchar *
truncate_escape_string (const gchar *str,
						int num_chars)
{
	gchar *escaped_str;

	if (!notify_supports_truncation && g_utf8_strlen (str, num_chars*2+1) > num_chars) {
		gchar *truncated_str;
		gchar *str2;

		/* allocate number of bytes and not number of utf-8 chars */
		str2 = g_malloc ((num_chars-1) * 2 * sizeof(gchar));

		g_utf8_strncpy (str2, str, num_chars-2);
		truncated_str = g_strdup_printf ("%s..", str2);
		escaped_str = g_markup_escape_text (truncated_str, strlen (truncated_str));
		g_free (str2);
		g_free (truncated_str);
	} else {
		escaped_str = g_markup_escape_text (str, strlen (str));
	}

	return escaped_str;
}

static gboolean
should_notify_unavailable (PurpleAccount *account)
{
	PurpleStatus *status;

	if (!purple_prefs_get_bool ("/plugins/gtk/libnotify/only_available"))
		return TRUE;

	status = purple_account_get_active_status (account);

	return purple_status_is_online (status) && purple_status_is_available (status);
}

static void
notify (const gchar *title,
		const gchar *body,
		PurpleBuddy *buddy,
		gboolean append)
{
	NotifyNotification *notification = NULL;
	GdkPixbuf *icon = NULL;
	PurpleBuddyIcon *buddy_icon = NULL;
	gchar *tr_body = NULL;
	PurpleContact *contact = NULL;

	if (buddy != NULL) {
		contact = purple_buddy_get_contact (buddy);
	}

	if (body)
		tr_body = truncate_escape_string (body, 60);
	else
		tr_body = NULL;

	/* If we're appending we shouldn't update an already
	   existing notification */
	if (!append && contact != NULL) {
		notification = g_hash_table_lookup (buddy_hash, contact);
	}

	if (notification != NULL) {
		if (notify_supports_append) {
			if (append) {
				notify_notification_set_hint_string(notification, "x-canonical-append", "allow");
			} else {
				notify_notification_set_hint_string(notification, "x-canonical-append", "off");
			}
		}

		notify_notification_update (notification, title, tr_body, NULL);

		/* this shouldn't be necessary, file a bug */
		notify_notification_show (notification, NULL);

		purple_debug_info (PLUGIN_ID, "notify(), update: "
						 "title: '%s', body: '%s', buddy: '%s'\n",
						 title, tr_body, buddy != NULL ? best_name (buddy) : "(null)");

		g_free (tr_body);
		return;
	}
	notification = notify_notification_new (title, tr_body, NULL, NULL);
	purple_debug_info (PLUGIN_ID, "notify(), new: "
					 "title: '%s', body: '%s', buddy: '%s'\n",
					 title, tr_body, buddy != NULL ? best_name (buddy) : "(null)");

	g_free (tr_body);

	if (notify_supports_append) {
		if (append) {
			notify_notification_set_hint_string(notification, "x-canonical-append", "allow");
		} else {
			notify_notification_set_hint_string(notification, "x-canonical-append", "off");
		}
	}

	if (buddy != NULL) {
		buddy_icon = purple_buddy_get_icon (buddy);
	}

	if (buddy_icon != NULL) {
		icon = pixbuf_from_buddy_icon (buddy_icon);
		purple_debug_info (PLUGIN_ID, "notify(), has a buddy icon.\n");
	} else {
		if (buddy != NULL) {
			icon = pidgin_create_prpl_icon (buddy->account, PIDGIN_PRPL_ICON_LARGE);
			purple_debug_info (PLUGIN_ID, "notify(), has a prpl icon.\n");
		}
	}

	if (icon != NULL) {
		notify_notification_set_icon_from_pixbuf (notification, icon);
		g_object_unref (icon);
	} else {
		purple_debug_warning (PLUGIN_ID, "notify(), couldn't find any icon!\n");
	}

	if (contact != NULL) {
		g_hash_table_insert (buddy_hash, contact, notification);
	}

	g_object_set_data (G_OBJECT(notification), "contact", contact);

	g_signal_connect (notification, "closed", G_CALLBACK(closed_cb), NULL);

	notify_notification_set_urgency (notification, NOTIFY_URGENCY_NORMAL);

	if (notify_supports_actions) {
		notify_notification_add_action (notification, "show", _("Show"), action_cb, NULL, NULL);
	}

	if (!notify_notification_show (notification, NULL)) {
		purple_debug_error (PLUGIN_ID, "notify(), failed to send notification\n");
	}

}

static void
notify_buddy_signon_cb (PurpleBuddy *buddy,
						gpointer data)
{
	gchar *tr_name;
	gboolean blocked;

	g_return_if_fail (buddy);

	if (!purple_prefs_get_bool ("/plugins/gtk/libnotify/signon"))
		return;

	if (!purple_account_is_connected(buddy->account))
		return;

	PurplePresence * account_pres = purple_account_get_presence(buddy->account);
	PurplePresence * buddy_pres = purple_buddy_get_presence(buddy);

	if (difftime(purple_presence_get_login_time(buddy_pres), purple_presence_get_login_time(account_pres)) < 30.0) {
		/* If the account came online at basically the same time as
		   the buddy it's really unlikely that's when the buddy came
		   online.  More likely that the initialization of the buddy
		   was related to the account coming online. */
		return;
	}

	blocked = purple_prefs_get_bool ("/plugins/gtk/libnotify/blocked");
	if (!purple_privacy_check (buddy->account, buddy->name) && blocked)
		return;

	if (!should_notify_unavailable (purple_buddy_get_account (buddy)))
		return;

	tr_name = truncate_escape_string (best_name (buddy), 25);

	notify (tr_name, _("is online"), buddy, FALSE);

	g_free (tr_name);
}

static void
notify_buddy_signoff_cb (PurpleBuddy *buddy,
						 gpointer data)
{
	gchar *tr_name;
	gboolean blocked;

	g_return_if_fail (buddy);

	if (!purple_prefs_get_bool ("/plugins/gtk/libnotify/signoff"))
		return;

	PurplePresence * account_pres = purple_account_get_presence(buddy->account);
	PurplePresence * buddy_pres = purple_buddy_get_presence(buddy);

	if (difftime(purple_presence_get_login_time(buddy_pres), purple_presence_get_login_time(account_pres)) < 30.0) {
		/* If the account came online at basically the same time as
		   the buddy it's really unlikely that's when the buddy came
		   online.  More likely that the initialization of the buddy
		   was related to the account coming online. */
		return;
	}

	blocked = purple_prefs_get_bool ("/plugins/gtk/libnotify/blocked");
	if (!purple_privacy_check (buddy->account, buddy->name) && blocked)
		return;

	if (!should_notify_unavailable (purple_buddy_get_account (buddy)))
		return;

	tr_name = truncate_escape_string (best_name (buddy), 25);

	notify (tr_name, _("is offline"), buddy, FALSE);

	g_free (tr_name);
}

static void
notify_msg_sent (PurpleAccount *account,
				 const gchar *sender,
				 const gchar *message,
				 PurpleConversation * conv)
{
	PurpleBuddy *buddy = NULL;
	gchar *body = NULL, *tr_name = NULL;
	gboolean blocked;

	blocked = purple_prefs_get_bool ("/plugins/gtk/libnotify/blocked");
	if (!purple_privacy_check(account, sender) && blocked)
		return;

	buddy = purple_find_buddy (account, sender);

	if (buddy != NULL) {
		tr_name = truncate_escape_string (best_name (buddy), 25);
	} else {
		if (conv != NULL) {
			const gchar * temp = purple_conversation_get_title(conv);
			if (temp != NULL) {
				if (sender == NULL || !g_strcmp0(sender, temp)) {
					tr_name = g_strdup(temp);
				} else {
					tr_name = g_strdup_printf("%s (%s)", sender, temp);
				}
			} else {
				if (sender != NULL) {
					tr_name = g_strdup(sender);
				}
			}
		}
	}

	if (tr_name == NULL) {
		purple_debug_warning(PLUGIN_ID, "Unable to find a title for the notification");
		return;
	}

	body = purple_markup_strip_html (message);

	notify (tr_name, body, buddy, TRUE);

	g_free (tr_name);
	g_free (body);
}

static void
notify_new_message_cb (PurpleAccount *account,
					   const gchar *sender,
					   const gchar *message,
					   int flags,
					   gpointer data)
{
	PurpleConversation *conv;

	if (!purple_prefs_get_bool ("/plugins/gtk/libnotify/newmsg"))
		return;

	conv = purple_find_conversation_with_account (PURPLE_CONV_TYPE_IM, sender, account);

#ifndef DEBUG /* in debug mode, always show notifications */
	if (conv && purple_conversation_has_focus (conv)) {
		purple_debug_info (PLUGIN_ID, "Conversation has focus 0x%x\n", GPOINTER_TO_UINT(conv));
		return;
	}
#endif

	if (conv && purple_prefs_get_bool ("/plugins/gtk/libnotify/newconvonly")) {
		purple_debug_info (PLUGIN_ID, "Conversation is not new 0x%x\n", GPOINTER_TO_UINT(conv));
		return;
	}

	if (conv == NULL) {
		purple_debug_warning(PLUGIN_ID, "Notify Message send has NULL Conversation, assuming hidden");
		return;
	}

	if (!should_notify_unavailable (account))
		return;

	PidginConversation * pconv = PIDGIN_CONVERSATION(conv);
	if (pconv == NULL) {
		purple_debug_warning(PLUGIN_ID, "Pidgin Conversation doesn't exist yet.");
		return;
	}
	if (pconv->entry == NULL || pconv->imhtml == NULL) {
		purple_debug_warning(PLUGIN_ID, "Pidgin Conversation doesn't have widgets yet.");
		return;
	}
	if (GTK_WIDGET_HAS_FOCUS(pconv->entry) || GTK_WIDGET_HAS_FOCUS(pconv->imhtml)) {
		purple_debug_warning(PLUGIN_ID, "Pidgin conversation's widgets are in focus");
		return;
	}

	notify_msg_sent (account, sender, message, conv);
}

static void
notify_chat_nick (PurpleAccount *account,
				  const gchar *sender,
				  const gchar *message,
				  PurpleConversation *conv,
				  gpointer data)
{
	gchar *nick;

	nick = (gchar *)purple_conv_chat_get_nick (PURPLE_CONV_CHAT(conv));
	if (nick && !strcmp (sender, nick))
		return;

	if (!purple_utf8_has_word (message, nick))
		return;

	PidginConversation * pconv = PIDGIN_CONVERSATION(conv);
	if (pconv == NULL) {
		purple_debug_warning(PLUGIN_ID, "Pidgin Conversation doesn't exist yet.");
		return;
	}
	if (pconv->entry == NULL || pconv->imhtml == NULL) {
		purple_debug_warning(PLUGIN_ID, "Pidgin Conversation doesn't have widgets yet.");
		return;
	}
	if (GTK_WIDGET_HAS_FOCUS(pconv->entry) || GTK_WIDGET_HAS_FOCUS(pconv->imhtml)) {
		purple_debug_warning(PLUGIN_ID, "Pidgin conversation's widgets are in focus");
		return;
	}

	notify_msg_sent (account, sender, message, conv);
}

static gboolean
indicate_focus_cb (GtkWidget * widget, GdkEventFocus * event, gpointer data)
{
	PurpleConversation * conv = (PurpleConversation *)data;
	IndicateIndicator * indicator = purple_conversation_get_data(conv, "indicate-indicator");
	PidginConversation * pconv = PIDGIN_CONVERSATION(conv);

	if (indicator == NULL) {
		return FALSE;
	}

	indicate_indicator_hide(indicator);
	g_object_unref(indicator);

	purple_conversation_set_data(conv, "indicate-indicator", NULL);
	
	g_signal_handlers_disconnect_by_func(G_OBJECT(pconv->entry),  indicate_focus_cb, conv);
	g_signal_handlers_disconnect_by_func(G_OBJECT(pconv->imhtml), indicate_focus_cb, conv);

	/* Don't swallow event */
	return FALSE;
}

/*
 * Taken from libtomboy, (C) 2008 Novell, LGPL v2 or later
 */

static void
tomboy_window_override_user_time (GtkWindow *window)
{
	guint32 ev_time = gtk_get_current_event_time();

	if (ev_time == 0) {
		/* 
		 * FIXME: Global keypresses use an event filter on the root
		 * window, which processes events before GDK sees them.
		 */
		//ev_time = tomboy_keybinder_get_current_event_time ();
    ev_time = GDK_CURRENT_TIME;
	}
	if (ev_time == 0) {
		gint ev_mask = gtk_widget_get_events (GTK_WIDGET(window));
		if (!(ev_mask & GDK_PROPERTY_CHANGE_MASK)) {
			gtk_widget_add_events (GTK_WIDGET (window),
					       GDK_PROPERTY_CHANGE_MASK);
		}

		/* 
		 * NOTE: Last resort for D-BUS or other non-interactive
		 *       openings.  Causes roundtrip to server.  Lame. 
		 */
		ev_time = gdk_x11_get_server_time (GTK_WIDGET(window)->window);
	}

	gdk_x11_window_set_user_time (GTK_WIDGET(window)->window, ev_time);
}

static void 
really_present_window (GtkWindow *window)
{
	if (!GTK_WIDGET_REALIZED (window))
		gtk_widget_realize (GTK_WIDGET (window));

	tomboy_window_override_user_time (window);

	gtk_window_present (window);
}

void
indicate_show_cb (IndicateIndicator * indicator, PurpleConversation * conv)
{
	/* g_debug ("indicate_show_cb()"); */

	if (conv == NULL) {
		purple_debug_warning(PLUGIN_ID, "\tNULL conversation");
	}

	PidginConversation * gtkconv = PIDGIN_CONVERSATION(conv);
	if (gtkconv == NULL) {
		purple_debug_warning(PLUGIN_ID, "\tNULL Pidgin Conversation");
	}

	pidgin_conv_switch_active_conversation(conv);
	pidgin_conv_window_switch_gtkconv(gtkconv->win, gtkconv);
	really_present_window(GTK_WINDOW(gtkconv->win->window));

	return;
}

static void
conv_delete_cb (PurpleConversation * conv, void * data)
{
	g_debug("Pidgin conv delete: %s", purple_conversation_get_name(conv));
	IndicateIndicator * indicator = INDICATE_INDICATOR(purple_conversation_get_data(conv, "indicate-indicator"));
	if (indicator != NULL) {
		indicate_indicator_hide(indicator);
		g_object_unref(G_OBJECT(indicator));
		purple_conversation_set_data(conv, "indicate-indicator", NULL);
	}

	PidginConversation * pconv = PIDGIN_CONVERSATION(conv);
	if (pconv != NULL) {
		g_signal_handlers_disconnect_by_func(G_OBJECT(pconv->entry),  G_CALLBACK(indicate_focus_cb), conv);
		g_signal_handlers_disconnect_by_func(G_OBJECT(pconv->imhtml), G_CALLBACK(indicate_focus_cb), conv);
	}

	return;
}

static void
indicate_chat_nick (PurpleAccount *account,
				    const gchar *sender,
				    const gchar *message,
				    PurpleConversation *conv,
				    gpointer data)
{
	PurpleBuddy *buddy = NULL;
	GdkPixbuf *icon = NULL;
	gchar *tr_name = NULL;
	PurpleBuddyIcon * buddy_icon = NULL;

	/* g_debug("Entering indicate_chat_nick"); */

	if (conv == NULL) {
		purple_debug_warning(PLUGIN_ID, "Conversation is NULL, not sure what to do with that!");
		return;
	}

	if (purple_conversation_get_type(conv) == PURPLE_CONV_TYPE_CHAT) {
		gchar * nick = (gchar *)purple_conv_chat_get_nick (PURPLE_CONV_CHAT(conv));
		if (nick && !strcmp (sender, nick))
			return;

		if (!purple_utf8_has_word (message, nick))
			return;
	}

	IndicateIndicatorMessage * indicator = purple_conversation_get_data(conv, "indicate-indicator");
	if (indicator != NULL) {
		/* We've already indicated this one, let's set the time */
		/* g_debug("Updating indicator time"); */
		GTimeVal time; g_get_current_time(&time);
		indicate_indicator_set_property_time(INDICATE_INDICATOR(indicator), "time", &time);
		return;
	}

	PidginConversation * pconv = PIDGIN_CONVERSATION(conv);
	if (pconv == NULL) {
		purple_debug_warning(PLUGIN_ID, "Pidgin Conversation doesn't exist yet.");
		return;
	}
	if (pconv->entry == NULL || pconv->imhtml == NULL) {
		purple_debug_warning(PLUGIN_ID, "Pidgin Conversation doesn't have widgets yet.");
		return;
	}
	if (GTK_WIDGET_HAS_FOCUS(pconv->entry) || GTK_WIDGET_HAS_FOCUS(pconv->imhtml)) {
		purple_debug_warning(PLUGIN_ID, "Pidgin conversation's widgets are in focus");
		return;
	}

	if (account != NULL && sender != NULL) {
		buddy = purple_find_buddy (account, sender);
	} else {
		purple_debug_warning(PLUGIN_ID, "We can't create an indicator for an account or a sender that don't exist!");
	}

	if (buddy != NULL) {
		purple_debug_warning(PLUGIN_ID, "Unable to find buddy.");

		buddy_icon = purple_buddy_get_icon(buddy);
	}

	if (buddy_icon != NULL) {
		icon = pixbuf_from_buddy_icon(buddy_icon);
	} else {
		if (buddy != NULL) {
			icon = pidgin_create_prpl_icon(buddy->account, PIDGIN_PRPL_ICON_LARGE);
		}
	}

	if (buddy != NULL) {
		tr_name = truncate_escape_string (best_name (buddy), 25);
	} else {
		const gchar * temp = purple_conversation_get_title(conv);
		if (temp != NULL) {
			if (sender == NULL || !g_strcmp0(sender, temp)) {
				tr_name = g_strdup(temp);
			} else {
				tr_name = g_strdup_printf("%s (%s)", sender, temp);
			}
		} else {
			if (sender != NULL) {
				tr_name = g_strdup(sender);
			}
		}
	}

	if (tr_name == NULL) {
		purple_debug_warning(PLUGIN_ID, "Unable to determine a sender");
		return;
	}

	indicator = indicate_indicator_message_new();

	indicate_indicator_set_property(INDICATE_INDICATOR(indicator), "subtype", "im");
	indicate_indicator_set_property(INDICATE_INDICATOR(indicator), "sender", tr_name);
	if (icon) {
		indicate_indicator_set_property_icon(INDICATE_INDICATOR(indicator), "icon", icon);
		g_object_unref(G_OBJECT(icon));
	}
	GTimeVal time; g_get_current_time(&time);
	indicate_indicator_set_property_time(INDICATE_INDICATOR(indicator), "time", &time);
	indicate_indicator_show(INDICATE_INDICATOR(indicator));

	purple_conversation_set_data(conv, "indicate-indicator", indicator);

	g_signal_connect(G_OBJECT(pconv->entry),  "focus-in-event", G_CALLBACK(indicate_focus_cb), conv);
	g_signal_connect(G_OBJECT(pconv->imhtml), "focus-in-event", G_CALLBACK(indicate_focus_cb), conv);

	g_signal_connect(G_OBJECT(indicator), INDICATE_INDICATOR_SIGNAL_DISPLAY, G_CALLBACK(indicate_show_cb), conv);

	return;
}

static void
indicate_new_message_cb (PurpleAccount *account,
					   const gchar *sender,
					   const gchar *message,
					   int flags,
					   gpointer data)
{
	PurpleConversation * conv = purple_find_conversation_with_account (PURPLE_CONV_TYPE_IM, sender, account);

	return indicate_chat_nick(account, sender, message, conv, data);
}

static void
indicate_server_display (IndicateServer * server, gpointer data)
{
	if (visibility_managed == FALSE) {
		pidgin_blist_visibility_manager_add();
		visibility_managed = TRUE;
	}

	pidgin_blist_toggle_visibility();
}

static void
notify_check_caps_helper (gpointer data, gpointer user_data)
{
	gchar * cap = (gchar *)data;

	if (cap == NULL) return;

	if (!strcmp(cap, "actions")) {
		notify_supports_actions = TRUE;
	} else if (!strcmp(cap, "append")) {
		notify_supports_append = TRUE;
	} else if (!strcmp(cap, "x-canonical-append")) {
		notify_supports_append = TRUE;
	} else if (!strcmp(cap, "truncation")) {
		notify_supports_truncation = TRUE;
	} else if (!strcmp(cap, "x-canonical-truncation")) {
		notify_supports_truncation = TRUE;
	}

	return;
}

static void
notify_check_caps(void)
{
	GList * caps = notify_get_server_caps();

	g_list_foreach(caps, notify_check_caps_helper, NULL);
	g_list_foreach(caps, (GFunc)g_free, NULL);
	g_list_free(caps);

	return;
}

static gboolean
plugin_load (PurplePlugin *plugin)
{
	void *conv_handle, *blist_handle, *conn_handle;

	if (!notify_is_initted () && !notify_init ("Pidgin")) {
		purple_debug_error (PLUGIN_ID, "libnotify not running!\n");
		return FALSE;
	}

	notify_check_caps();

	indicate_server = indicate_server_ref_default();
	indicate_server_set_type(indicate_server, "message.instant");
	indicate_server_set_desktop_file(indicate_server, "/usr/share/applications/pidgin.desktop");
	g_signal_connect(G_OBJECT(indicate_server), INDICATE_SERVER_SIGNAL_SERVER_DISPLAY, G_CALLBACK(indicate_server_display), NULL);
	indicate_server_show(indicate_server);

	conv_handle = purple_conversations_get_handle ();
	blist_handle = purple_blist_get_handle ();
	conn_handle = purple_connections_get_handle();

	buddy_hash = g_hash_table_new (NULL, NULL);

	purple_signal_connect (blist_handle, "buddy-signed-on", plugin,
						PURPLE_CALLBACK(notify_buddy_signon_cb), NULL);

	purple_signal_connect (blist_handle, "buddy-signed-off", plugin,
						PURPLE_CALLBACK(notify_buddy_signoff_cb), NULL);

	purple_signal_connect (conv_handle, "received-im-msg", plugin,
						PURPLE_CALLBACK(notify_new_message_cb), NULL);

	purple_signal_connect (conv_handle, "received-chat-msg", plugin,
						PURPLE_CALLBACK(notify_chat_nick), NULL);

	purple_signal_connect (conv_handle, "received-im-msg", plugin,
						PURPLE_CALLBACK(indicate_new_message_cb), NULL);

	purple_signal_connect (conv_handle, "received-chat-msg", plugin,
						PURPLE_CALLBACK(indicate_chat_nick), NULL);

	purple_signal_connect (conv_handle, "deleting-conversation", plugin,
	                    PURPLE_CALLBACK(conv_delete_cb), NULL);

	return TRUE;
}

static gboolean
plugin_unload (PurplePlugin *plugin)
{
	void *conv_handle, *blist_handle, *conn_handle;

	conv_handle = purple_conversations_get_handle ();
	blist_handle = purple_blist_get_handle ();
	conn_handle = purple_connections_get_handle();

	purple_signal_disconnect (blist_handle, "buddy-signed-on", plugin,
							PURPLE_CALLBACK(notify_buddy_signon_cb));

	purple_signal_disconnect (blist_handle, "buddy-signed-off", plugin,
							PURPLE_CALLBACK(notify_buddy_signoff_cb));

	purple_signal_disconnect (conv_handle, "received-im-msg", plugin,
							PURPLE_CALLBACK(notify_new_message_cb));

	purple_signal_disconnect (conv_handle, "received-chat-msg", plugin,
							PURPLE_CALLBACK(notify_chat_nick));

	purple_signal_disconnect (conv_handle, "received-im-msg", plugin,
							PURPLE_CALLBACK(indicate_new_message_cb));

	purple_signal_disconnect (conv_handle, "received-chat-msg", plugin,
							PURPLE_CALLBACK(indicate_chat_nick));

	purple_signal_disconnect (conv_handle, "deleting-conversation", plugin,
							PURPLE_CALLBACK(conv_delete_cb));

	g_hash_table_destroy (buddy_hash);

	notify_uninit ();

	indicate_server_hide(indicate_server);

	if (visibility_managed == TRUE) {
		pidgin_blist_visibility_manager_remove();
	}

	g_object_unref(G_OBJECT(indicate_server));

	return TRUE;
}

static PurplePluginUiInfo prefs_info = {
    get_plugin_pref_frame,
    0,						/* page num (Reserved) */
    NULL					/* frame (Reserved) */
};

static PurplePluginInfo info = {
    PURPLE_PLUGIN_MAGIC,										/* api version */
    PURPLE_MAJOR_VERSION,
    PURPLE_MINOR_VERSION,
    PURPLE_PLUGIN_STANDARD,									/* type */
    0,														/* ui requirement */
    0,														/* flags */
    NULL,													/* dependencies */
    PURPLE_PRIORITY_DEFAULT,									/* priority */
    
    PLUGIN_ID,												/* id */
    NULL,													/* name */
    VERSION,												/* version */
    NULL,													/* summary */
    NULL,													/* description */
    
    "Duarte Henriques <duarte.henriques@gmail.com>",		/* author */
    "http://sourceforge.net/projects/gaim-libnotify/",		/* homepage */
    
    plugin_load,			/* load */
    plugin_unload,			/* unload */
    NULL,					/* destroy */
    NULL,					/* ui info */
    NULL,					/* extra info */
    &prefs_info				/* prefs info */
};

static void
init_plugin (PurplePlugin *plugin)
{
	bindtextdomain (PACKAGE, LOCALEDIR);
	bind_textdomain_codeset (PACKAGE, "UTF-8");

	info.name = _("Libnotify Popups");
	info.summary = _("Displays popups via libnotify.");
	info.description = _("Pidgin-libnotify:\nDisplays popups via libnotify.");

	purple_prefs_add_none ("/plugins/gtk/libnotify");
	purple_prefs_add_bool ("/plugins/gtk/libnotify/newmsg", TRUE);
	purple_prefs_add_bool ("/plugins/gtk/libnotify/blocked", TRUE);
	purple_prefs_add_bool ("/plugins/gtk/libnotify/newconvonly", FALSE);
	purple_prefs_add_bool ("/plugins/gtk/libnotify/signon", TRUE);
	purple_prefs_add_bool ("/plugins/gtk/libnotify/signoff", FALSE);
	purple_prefs_add_bool ("/plugins/gtk/libnotify/only_available", FALSE);
}

PURPLE_INIT_PLUGIN(notify, init_plugin, info)

